@@ -1,23 +1,30 @@
-from dogpile.cache.api import CacheBackend, NO_VALUE
-from dogpile.cache import register_backend, CacheRegion
-from dogpile.cache.region import _backend_loader
-from . import eq_, assert_raises_message
-import itertools
-import time
-import pytest
-from threading import Thread, Lock
-from unittest import TestCase
-import random
 import collections
+import itertools
+import json
+import random
+from threading import Lock
+from threading import Thread
+import time
+from unittest import TestCase
+
+import pytest
+
+from dogpile.cache import CacheRegion
+from dogpile.cache import register_backend
+from dogpile.cache.api import CacheBackend
+from dogpile.cache.api import CacheMutex
+from dogpile.cache.api import NO_VALUE
+from dogpile.cache.region import _backend_loader
+from . import assert_raises_message
+from . import eq_
 
 
 class _GenericBackendFixture(object):
-
     @classmethod
     def setup_class(cls):
         backend_cls = _backend_loader.load(cls.backend)
         try:
-            arguments = cls.config_args.get('arguments', {})
+            arguments = cls.config_args.get("arguments", {})
             backend = backend_cls(arguments)
         except ImportError:
             pytest.skip("Backend %s not installed" % cls.backend)
@@ -37,6 +44,7 @@ class _GenericBackendFixture(object):
 
     region_args = {}
     config_args = {}
+    extra_arguments = {}
 
     _region_inst = None
     _backend_inst = None
@@ -44,7 +52,15 @@ class _GenericBackendFixture(object):
     _keys = set()
 
     def _region(self, backend=None, region_args={}, config_args={}):
-        _region_args = self.region_args.copy()
+        _region_args = {}
+
+        # TODO: maybe we use a class-level naming convention instead
+        # of a dict here so that arguments merge naturally
+
+        for cls in reversed(self.__class__.__mro__):
+            if "region_args" in cls.__dict__:
+                _region_args.update(cls.__dict__["region_args"])
+
         _region_args.update(**region_args)
         _config_args = self.config_args.copy()
         _config_args.update(config_args)
@@ -54,6 +70,7 @@ class _GenericBackendFixture(object):
                 key = existing_key_mangler(key)
             self._keys.add(key)
             return key
+
         self._region_inst = reg = CacheRegion(**_region_args)
 
         existing_key_mangler = self._region_inst.key_mangler
@@ -66,16 +83,16 @@ class _GenericBackendFixture(object):
     def _backend(self):
         backend_cls = _backend_loader.load(self.backend)
         _config_args = self.config_args.copy()
-        arguments = _config_args.get('arguments', {})
+        arguments = _config_args.get("arguments", {})
+        arguments = {**arguments, **self.extra_arguments}
         self._backend_inst = backend_cls(arguments)
         return self._backend_inst
 
 
 class _GenericBackendTest(_GenericBackendFixture, TestCase):
-
     def test_backend_get_nothing(self):
         backend = self._backend()
-        eq_(backend.get("some_key"), NO_VALUE)
+        eq_(backend.get_serialized("some_key"), NO_VALUE)
 
     def test_backend_delete_nothing(self):
         backend = self._backend()
@@ -83,14 +100,14 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
 
     def test_backend_set_get_value(self):
         backend = self._backend()
-        backend.set("some_key", "some value")
-        eq_(backend.get("some_key"), "some value")
+        backend.set_serialized("some_key", b"some value")
+        eq_(backend.get_serialized("some_key"), b"some value")
 
     def test_backend_delete(self):
         backend = self._backend()
-        backend.set("some_key", "some value")
+        backend.set_serialized("some_key", b"some value")
         backend.delete("some_key")
-        eq_(backend.get("some_key"), NO_VALUE)
+        eq_(backend.get_serialized("some_key"), NO_VALUE)
 
     def test_region_set_get_value(self):
         reg = self._region()
@@ -99,11 +116,11 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
 
     def test_region_set_multiple_values(self):
         reg = self._region()
-        values = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'}
+        values = {"key1": "value1", "key2": "value2", "key3": "value3"}
         reg.set_multi(values)
-        eq_(values['key1'], reg.get('key1'))
-        eq_(values['key2'], reg.get('key2'))
-        eq_(values['key3'], reg.get('key3'))
+        eq_(values["key1"], reg.get("key1"))
+        eq_(values["key2"], reg.get("key2"))
+        eq_(values["key3"], reg.get("key3"))
 
     def test_region_get_zero_multiple_values(self):
         reg = self._region()
@@ -121,35 +138,34 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
     def test_region_get_or_create_multi_w_should_cache_none(self):
         reg = self._region()
         values = reg.get_or_create_multi(
-            ['key1', 'key2', 'key3'], lambda *k: [None, None, None],
-            should_cache_fn=lambda v: v is not None)
+            ["key1", "key2", "key3"],
+            lambda *k: [None, None, None],
+            should_cache_fn=lambda v: v is not None,
+        )
         eq_(values, [None, None, None])
 
     def test_region_get_multiple_values(self):
         reg = self._region()
-        key1 = 'value1'
-        key2 = 'value2'
-        key3 = 'value3'
-        reg.set('key1', key1)
-        reg.set('key2', key2)
-        reg.set('key3', key3)
-        values = reg.get_multi(['key1', 'key2', 'key3'])
-        eq_(
-            [key1, key2, key3], values
-        )
+        key1 = "value1"
+        key2 = "value2"
+        key3 = "value3"
+        reg.set("key1", key1)
+        reg.set("key2", key2)
+        reg.set("key3", key3)
+        values = reg.get_multi(["key1", "key2", "key3"])
+        eq_([key1, key2, key3], values)
 
     def test_region_get_nothing_multiple(self):
         reg = self._region()
-        reg.delete_multi(['key1', 'key2', 'key3', 'key4', 'key5'])
-        values = {'key1': 'value1', 'key3': 'value3', 'key5': 'value5'}
+        reg.delete_multi(["key1", "key2", "key3", "key4", "key5"])
+        values = {"key1": "value1", "key3": "value3", "key5": "value5"}
         reg.set_multi(values)
         reg_values = reg.get_multi(
-            ['key1', 'key2', 'key3', 'key4', 'key5', 'key6'])
+            ["key1", "key2", "key3", "key4", "key5", "key6"]
+        )
         eq_(
             reg_values,
-            ["value1", NO_VALUE, "value3", NO_VALUE,
-                "value5", NO_VALUE
-             ]
+            ["value1", NO_VALUE, "value3", NO_VALUE, "value5", NO_VALUE],
         )
 
     def test_region_get_empty_multiple(self):
@@ -159,13 +175,13 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
 
     def test_region_delete_multiple(self):
         reg = self._region()
-        values = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'}
+        values = {"key1": "value1", "key2": "value2", "key3": "value3"}
         reg.set_multi(values)
-        reg.delete_multi(['key2', 'key10'])
-        eq_(values['key1'], reg.get('key1'))
-        eq_(NO_VALUE, reg.get('key2'))
-        eq_(values['key3'], reg.get('key3'))
-        eq_(NO_VALUE, reg.get('key10'))
+        reg.delete_multi(["key2", "key10"])
+        eq_(values["key1"], reg.get("key1"))
+        eq_(NO_VALUE, reg.get("key2"))
+        eq_(values["key3"], reg.get("key3"))
+        eq_(NO_VALUE, reg.get("key10"))
 
     def test_region_set_get_nothing(self):
         reg = self._region()
@@ -177,20 +193,22 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
 
         def creator():
             return "some value"
+
         eq_(reg.get_or_create("some key", creator), "some value")
 
+    @pytest.mark.time_intensive
     def test_threaded_dogpile(self):
         # run a basic dogpile concurrency test.
         # note the concurrency of dogpile itself
         # is intensively tested as part of dogpile.
-        reg = self._region(config_args={"expiration_time": .25})
+        reg = self._region(config_args={"expiration_time": 0.25})
         lock = Lock()
         canary = []
 
         def creator():
             ack = lock.acquire(False)
             canary.append(ack)
-            time.sleep(.25)
+            time.sleep(0.25)
             if ack:
                 lock.release()
             return "some value"
@@ -198,7 +216,7 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
         def f():
             for x in range(5):
                 reg.get_or_create("some key", creator)
-                time.sleep(.5)
+                time.sleep(0.5)
 
         threads = [Thread(target=f) for i in range(10)]
         for t in threads:
@@ -208,11 +226,10 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
         assert len(canary) > 2
         if not reg.backend.has_lock_timeout():
             assert False not in canary
-        else:
-            assert False in canary
 
+    @pytest.mark.time_intensive
     def test_threaded_get_multi(self):
-        reg = self._region(config_args={"expiration_time": .25})
+        reg = self._region(config_args={"expiration_time": 0.25})
         locks = dict((str(i), Lock()) for i in range(11))
 
         canary = collections.defaultdict(list)
@@ -230,7 +247,7 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
             for acq, key in zip(ack, keys):
                 canary[key].append(acq)
 
-            time.sleep(.5)
+            time.sleep(0.5)
 
             for acq, key in zip(ack, keys):
                 if acq:
@@ -240,12 +257,16 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
         def f():
             for x in range(5):
                 reg.get_or_create_multi(
-                    [str(random.randint(1, 10))
-                        for i in range(random.randint(1, 5))],
-                    creator)
-                time.sleep(.5)
+                    [
+                        str(random.randint(1, 10))
+                        for i in range(random.randint(1, 5))
+                    ],
+                    creator,
+                )
+                time.sleep(0.5)
+
         f()
-        return
+
         threads = [Thread(target=f) for i in range(5)]
         for t in threads:
             t.start()
@@ -263,16 +284,29 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
         reg.delete("some key")
         eq_(reg.get("some key"), NO_VALUE)
 
+    @pytest.mark.time_intensive
     def test_region_expire(self):
-        reg = self._region(config_args={"expiration_time": .25})
+        # TODO: ideally tests like these would not be using actual
+        # time(); instead, an artificial function where the increment
+        # can be controlled would be preferred.  this way tests need not
+        # have any delay in running and additionally there is no issue
+        # with very slow processing missing a timeout, as is often the
+        # case with this particular test
+
+        reg = self._region(config_args={"expiration_time": 0.75})
         counter = itertools.count(1)
 
         def creator():
             return "some value %d" % next(counter)
+
         eq_(reg.get_or_create("some key", creator), "some value 1")
-        time.sleep(.4)
+        time.sleep(0.85)
+        # expiration is definitely hit
         eq_(reg.get("some key", ignore_expiration=True), "some value 1")
         eq_(reg.get_or_create("some key", creator), "some value 2")
+
+        # this line needs to run less the .75 sec before the previous
+        # two or it hits the expiration
         eq_(reg.get("some key"), "some value 2")
 
     def test_decorated_fn_functionality(self):
@@ -307,14 +341,48 @@ class _GenericBackendTest(_GenericBackendFixture, TestCase):
             raise Exception("boom")
 
         assert_raises_message(
-            Exception,
-            "boom",
-            reg.get_or_create, "some_key", boom
+            Exception, "boom", reg.get_or_create, "some_key", boom
         )
 
 
-class _GenericMutexTest(_GenericBackendFixture, TestCase):
+class _GenericSerializerTest(TestCase):
+    # Inheriting from this class will make test cases
+    # use these serialization arguments
+    region_args = {
+        "serializer": lambda v: json.dumps(v).encode("ascii"),
+        "deserializer": json.loads,
+    }
 
+    def test_uses_serializer(self):
+        region = self._region()
+
+        backend = region.backend
+        value = {"foo": ["bar", 1, False, None]}
+        region.set("k", value)
+
+        raw = backend.get_serialized("k")
+
+        assert isinstance(raw, bytes)
+
+        pipe = raw.find(b"|")
+        payload = raw[pipe + 1 :]
+        eq_(payload, self.region_args["serializer"](value))
+        eq_(region._parse_serialized_from_backend(raw).payload, value)
+
+    def test_uses_deserializer(self):
+        region = self._region()
+
+        value = {"foo": ["bar", 1, False, None]}
+        region.set("k", value)
+
+        asserted = region.get("k")
+
+        eq_(asserted, value)
+
+    # TODO: test set_multi, get_multi
+
+
+class _GenericMutexTest(_GenericBackendFixture, TestCase):
     def test_mutex(self):
         backend = self._backend()
         mutex = backend.get_mutex("foo")
@@ -328,6 +396,13 @@ class _GenericMutexTest(_GenericBackendFixture, TestCase):
         assert ac3
         mutex.release()
 
+    def test_subclass_match(self):
+        backend = self._backend()
+        mutex = backend.get_mutex("foo")
+
+        assert isinstance(mutex, CacheMutex)
+
+    @pytest.mark.time_intensive
     def test_mutex_threaded(self):
         backend = self._backend()
         backend.get_mutex("foo")
@@ -342,11 +417,11 @@ class _GenericMutexTest(_GenericBackendFixture, TestCase):
                 for y in range(5):
                     ack = lock.acquire(False)
                     canary.append(ack)
-                    time.sleep(.002)
+                    time.sleep(0.002)
                     if ack:
                         lock.release()
                 mutex.release()
-                time.sleep(.02)
+                time.sleep(0.02)
 
         threads = [Thread(target=f) for i in range(5)]
         for t in threads:
@@ -381,18 +456,11 @@ class _GenericMutexTest(_GenericBackendFixture, TestCase):
         def create_bar():
             return "bar"
 
-        eq_(
-            reg.get_or_create("foo", create_foo),
-            "foobar"
-        )
-        eq_(
-            reg.get_or_create("foo", create_foo),
-            "foobar"
-        )
+        eq_(reg.get_or_create("foo", create_foo), "foobar")
+        eq_(reg.get_or_create("foo", create_foo), "foobar")
 
 
 class MockMutex(object):
-
     def __init__(self, key):
         self.key = key
 
@@ -404,7 +472,6 @@ class MockMutex(object):
 
 
 class MockBackend(CacheBackend):
-
     def __init__(self, arguments):
         self.arguments = arguments
         self._cache = {}
@@ -419,9 +486,7 @@ class MockBackend(CacheBackend):
             return NO_VALUE
 
     def get_multi(self, keys):
-        return [
-            self.get(key) for key in keys
-        ]
+        return [self.get(key) for key in keys]
 
     def set(self, key, value):
         self._cache[key] = value
@@ -436,4 +501,6 @@ class MockBackend(CacheBackend):
     def delete_multi(self, keys):
         for key in keys:
             self.delete(key)
+
+
 register_backend("mock", __name__, "MockBackend")
