@@ -1,15 +1,27 @@
+from collections import defaultdict
+import configparser
+import datetime
+import io
+import itertools
+import time
+from unittest import mock
 from unittest import TestCase
-from dogpile.cache.api import NO_VALUE, CachedValue
+
+from dogpile.cache import CacheRegion
 from dogpile.cache import exception
-from dogpile.cache import make_region, CacheRegion
+from dogpile.cache import make_region
+from dogpile.cache import util
+from dogpile.cache.api import CacheBackend
+from dogpile.cache.api import CachedValue
+from dogpile.cache.api import CacheMutex
+from dogpile.cache.api import NO_VALUE
 from dogpile.cache.proxy import ProxyBackend
 from dogpile.cache.region import _backend_loader
 from dogpile.cache.region import RegionInvalidationStrategy
-from . import eq_, is_, assert_raises_message, io, configparser
-import time
-import datetime
-import itertools
-from collections import defaultdict
+from dogpile.cache.region import value_version
+from . import assert_raises_message
+from . import eq_
+from . import is_
 from ._fixtures import MockBackend
 
 
@@ -23,54 +35,56 @@ class APITest(TestCase):
 
 
 class RegionTest(TestCase):
-
     def _region(self, init_args={}, config_args={}, backend="mock"):
         reg = CacheRegion(**init_args)
         reg.configure(backend, **config_args)
         return reg
 
     def test_set_name(self):
-        my_region = make_region(name='my-name')
-        eq_(my_region.name, 'my-name')
+        my_region = make_region(name="my-name")
+        eq_(my_region.name, "my-name")
 
     def test_instance_from_dict(self):
         my_conf = {
-            'cache.example.backend': 'mock',
-            'cache.example.expiration_time': 600,
-            'cache.example.arguments.url': '127.0.0.1'
+            "cache.example.backend": "mock",
+            "cache.example.expiration_time": 600,
+            "cache.example.arguments.url": "127.0.0.1",
         }
         my_region = make_region()
-        my_region.configure_from_config(my_conf, 'cache.example.')
+        my_region.configure_from_config(my_conf, "cache.example.")
         eq_(my_region.expiration_time, 600)
         assert isinstance(my_region.backend, MockBackend) is True
-        eq_(my_region.backend.arguments, {'url': '127.0.0.1'})
+        eq_(my_region.backend.arguments, {"url": "127.0.0.1"})
 
     def test_instance_from_config_string(self):
-        my_conf = \
-            '[xyz]\n'\
-            'cache.example.backend=mock\n'\
-            'cache.example.expiration_time=600\n'\
-            'cache.example.arguments.url=127.0.0.1\n'\
-            'cache.example.arguments.dogpile_lockfile=false\n'\
-            'cache.example.arguments.xyz=None\n'
+        my_conf = (
+            "[xyz]\n"
+            "cache.example.backend=mock\n"
+            "cache.example.expiration_time=600\n"
+            "cache.example.arguments.url=127.0.0.1\n"
+            "cache.example.arguments.dogpile_lockfile=false\n"
+            "cache.example.arguments.xyz=None\n"
+        )
 
         my_region = make_region()
         config = configparser.ConfigParser()
-        config.readfp(io.StringIO(my_conf))
+
+        config.read_file(io.StringIO(my_conf))
 
         my_region.configure_from_config(
-            dict(config.items('xyz')), 'cache.example.')
+            dict(config.items("xyz")), "cache.example."
+        )
         eq_(my_region.expiration_time, 600)
         assert isinstance(my_region.backend, MockBackend) is True
-        eq_(my_region.backend.arguments, {
-            'url': '127.0.0.1',
-            'dogpile_lockfile': False, 'xyz': None})
+        eq_(
+            my_region.backend.arguments,
+            {"url": "127.0.0.1", "dogpile_lockfile": False, "xyz": None},
+        )
 
     def test_datetime_expiration_time(self):
         my_region = make_region()
         my_region.configure(
-            backend='mock',
-            expiration_time=datetime.timedelta(days=1, hours=8)
+            backend="mock", expiration_time=datetime.timedelta(days=1, hours=8)
         )
         eq_(my_region.expiration_time, 32 * 60 * 60)
 
@@ -80,7 +94,9 @@ class RegionTest(TestCase):
         assert_raises_message(
             exception.ValidationError,
             "expiration_time is not a number or timedelta.",
-            my_region.configure, 'mock', 'one hour'
+            my_region.configure,
+            "mock",
+            "one hour",
         )
 
     def test_key_mangler_argument(self):
@@ -103,7 +119,8 @@ class RegionTest(TestCase):
         eq_(reg.get("some key"), "some value")
         eq_(
             reg.get_or_create("some key", lambda: "some new value"),
-            "some value")
+            "some value",
+        )
         reg.delete("some key")
         eq_(reg.get("some key"), NO_VALUE)
 
@@ -113,7 +130,8 @@ class RegionTest(TestCase):
         assert_raises_message(
             exception.RegionAlreadyConfigured,
             "This region is already configured",
-            reg.configure, "mock"
+            reg.configure,
+            "mock",
         )
         eq_(reg.is_configured, True)
 
@@ -148,7 +166,9 @@ class RegionTest(TestCase):
         assert_raises_message(
             exception.RegionNotConfigured,
             "No backend is configured on this region.",
-            getattr, reg, "backend"
+            getattr,
+            reg,
+            "backend",
         )
         eq_(reg.is_configured, False)
 
@@ -157,7 +177,8 @@ class RegionTest(TestCase):
         assert_raises_message(
             exception.PluginNotFound,
             "Couldn't find cache plugin to load: unknown",
-            reg.configure, 'unknown'
+            reg.configure,
+            "unknown",
         )
         eq_(reg.is_configured, False)
 
@@ -178,6 +199,7 @@ class RegionTest(TestCase):
 
         def creator():
             return "some value"
+
         eq_(reg.get_or_create("some key", creator), "some value")
 
     def test_multi_creator(self):
@@ -185,8 +207,11 @@ class RegionTest(TestCase):
 
         def creator(*keys):
             return ["some value %s" % key for key in keys]
-        eq_(reg.get_or_create_multi(["k3", "k2", "k5"], creator),
-            ['some value k3', 'some value k2', 'some value k5'])
+
+        eq_(
+            reg.get_or_create_multi(["k3", "k2", "k5"], creator),
+            ["some value k3", "some value k2", "some value k5"],
+        )
 
     def test_remove(self):
         reg = self._region()
@@ -201,12 +226,15 @@ class RegionTest(TestCase):
 
         def creator():
             return "some value %d" % next(counter)
+
         eq_(reg.get_or_create("some key", creator), "some value 1")
         time.sleep(2)
         is_(reg.get("some key"), NO_VALUE)
         eq_(reg.get("some key", ignore_expiration=True), "some value 1")
-        eq_(reg.get_or_create("some key", creator, expiration_time=-1),
-            "some value 1")
+        eq_(
+            reg.get_or_create("some key", creator, expiration_time=-1),
+            "some value 1",
+        )
         eq_(reg.get_or_create("some key", creator), "some value 2")
         eq_(reg.get("some key"), "some value 2")
 
@@ -216,33 +244,40 @@ class RegionTest(TestCase):
 
         def creator(*keys):
             return ["some value %s %d" % (key, next(counter)) for key in keys]
-        eq_(reg.get_or_create_multi(["k3", "k2", "k5"], creator),
-            ['some value k3 2', 'some value k2 1', 'some value k5 3'])
+
+        eq_(
+            reg.get_or_create_multi(["k3", "k2", "k5"], creator),
+            ["some value k3 2", "some value k2 1", "some value k5 3"],
+        )
         time.sleep(2)
         is_(reg.get("k2"), NO_VALUE)
         eq_(reg.get("k2", ignore_expiration=True), "some value k2 1")
-        eq_(reg.get_or_create_multi(["k3", "k2"], creator, expiration_time=-1),
-            ['some value k3 2', 'some value k2 1'])
-        eq_(reg.get_or_create_multi(["k3", "k2"], creator),
-            ['some value k3 5', 'some value k2 4'])
+        eq_(
+            reg.get_or_create_multi(["k3", "k2"], creator, expiration_time=-1),
+            ["some value k3 2", "some value k2 1"],
+        )
+        eq_(
+            reg.get_or_create_multi(["k3", "k2"], creator),
+            ["some value k3 5", "some value k2 4"],
+        )
         eq_(reg.get("k2"), "some value k2 4")
 
     def test_expire_on_get(self):
-        reg = self._region(config_args={"expiration_time": .5})
+        reg = self._region(config_args={"expiration_time": 0.5})
         reg.set("some key", "some value")
         eq_(reg.get("some key"), "some value")
         time.sleep(1)
         is_(reg.get("some key"), NO_VALUE)
 
     def test_ignore_expire_on_get(self):
-        reg = self._region(config_args={"expiration_time": .5})
+        reg = self._region(config_args={"expiration_time": 0.5})
         reg.set("some key", "some value")
         eq_(reg.get("some key"), "some value")
         time.sleep(1)
         eq_(reg.get("some key", ignore_expiration=True), "some value")
 
     def test_override_expire_on_get(self):
-        reg = self._region(config_args={"expiration_time": .5})
+        reg = self._region(config_args={"expiration_time": 0.5})
         reg.set("some key", "some value")
         eq_(reg.get("some key"), "some value")
         time.sleep(1)
@@ -255,18 +290,23 @@ class RegionTest(TestCase):
 
         def creator():
             return "some value %d" % next(counter)
-        eq_(reg.get_or_create("some key", creator, expiration_time=1),
-            "some value 1")
+
+        eq_(
+            reg.get_or_create("some key", creator, expiration_time=1),
+            "some value 1",
+        )
         time.sleep(2)
         eq_(reg.get("some key"), "some value 1")
-        eq_(reg.get_or_create("some key", creator, expiration_time=1),
-            "some value 2")
+        eq_(
+            reg.get_or_create("some key", creator, expiration_time=1),
+            "some value 2",
+        )
         eq_(reg.get("some key"), "some value 2")
 
     def test_hard_invalidate_get(self):
         reg = self._region()
         reg.set("some key", "some value")
-        time.sleep(.1)
+        time.sleep(0.1)
         reg.invalidate()
         is_(reg.get("some key"), NO_VALUE)
 
@@ -276,18 +316,59 @@ class RegionTest(TestCase):
 
         def creator():
             return "some value %d" % next(counter)
-        eq_(reg.get_or_create("some key", creator),
-            "some value 1")
 
-        time.sleep(.1)
+        eq_(reg.get_or_create("some key", creator), "some value 1")
+
+        time.sleep(0.1)
         reg.invalidate()
-        eq_(reg.get_or_create("some key", creator),
-            "some value 2")
+        eq_(reg.get_or_create("some key", creator), "some value 2")
+
+        eq_(reg.get_or_create("some key", creator), "some value 2")
+
+        reg.invalidate()
+        eq_(reg.get_or_create("some key", creator), "some value 3")
+
+        eq_(reg.get_or_create("some key", creator), "some value 3")
+
+    def test_hard_invalidate_get_or_create_multi(self):
+        reg = self._region()
+        counter = itertools.count(1)
+
+        def creator(*keys):
+            return ["some value %s %d" % (k, next(counter)) for k in keys]
+
+        eq_(
+            reg.get_or_create_multi(["k1", "k2"], creator),
+            ["some value k1 1", "some value k2 2"],
+        )
+
+        time.sleep(0.1)
+        reg.invalidate()
+        eq_(
+            reg.get_or_create_multi(["k1", "k2"], creator),
+            ["some value k1 3", "some value k2 4"],
+        )
+
+        eq_(
+            reg.get_or_create_multi(["k1", "k2"], creator),
+            ["some value k1 3", "some value k2 4"],
+        )
+
+        reg.invalidate()
+        eq_(
+            reg.get_or_create_multi(["k1", "k2"], creator),
+            ["some value k1 5", "some value k2 6"],
+        )
+
+        eq_(
+            reg.get_or_create_multi(["k1", "k2"], creator),
+            ["some value k1 5", "some value k2 6"],
+        )
 
     def test_soft_invalidate_get(self):
         reg = self._region(config_args={"expiration_time": 1})
         reg.set("some key", "some value")
-        time.sleep(.1)
+        time.sleep(0.1)
         reg.invalidate(hard=False)
         is_(reg.get("some key"), NO_VALUE)
 
@@ -297,13 +378,12 @@ class RegionTest(TestCase):
 
         def creator():
             return "some value %d" % next(counter)
-        eq_(reg.get_or_create("some key", creator),
-            "some value 1")
 
-        time.sleep(.1)
+        eq_(reg.get_or_create("some key", creator), "some value 1")
+
+        time.sleep(0.1)
         reg.invalidate(hard=False)
-        eq_(reg.get_or_create("some key", creator),
-            "some value 2")
+        eq_(reg.get_or_create("some key", creator), "some value 2")
 
     def test_soft_invalidate_get_or_create_multi(self):
         reg = self._region(config_args={"expiration_time": 5})
@@ -312,31 +392,61 @@ class RegionTest(TestCase):
         def creator(*keys):
             v = values.pop(0)
             return [v for k in keys]
-        ret = reg.get_or_create_multi(
-            [1, 2], creator)
+
+        ret = reg.get_or_create_multi([1, 2], creator)
         eq_(ret, [1, 1])
-        time.sleep(.1)
+        time.sleep(0.1)
         reg.invalidate(hard=False)
-        ret = reg.get_or_create_multi(
-            [1, 2], creator)
+        ret = reg.get_or_create_multi([1, 2], creator)
         eq_(ret, [2, 2])
 
-    def test_soft_invalidate_requires_expire_time_get(self):
+    def test_soft_invalidate_requires_expire_time_get_first_call(self):
         reg = self._region()
+
         reg.invalidate(hard=False)
         assert_raises_message(
             exception.DogpileCacheException,
             "Non-None expiration time required for soft invalidation",
-            reg.get_or_create, "some key", lambda: "x"
+            reg.get_or_create,
+            "some key",
+            lambda: "x",
         )
 
-    def test_soft_invalidate_requires_expire_time_get_multi(self):
+    def test_soft_invalidate_requires_expire_time_get_second_call(self):
+        reg = self._region()
+        reg.get_or_create("some key", lambda: "x")
+
+        reg.invalidate(hard=False)
+        assert_raises_message(
+            exception.DogpileCacheException,
+            "Non-None expiration time required for soft invalidation",
+            reg.get_or_create,
+            "some key",
+            lambda: "x",
+        )
+
+    def test_soft_invalidate_requires_expire_time_get_multi_first_call(self):
         reg = self._region()
         reg.invalidate(hard=False)
         assert_raises_message(
             exception.DogpileCacheException,
             "Non-None expiration time required for soft invalidation",
-            reg.get_or_create_multi, ["k1", "k2"], lambda k: "x"
+            reg.get_or_create_multi,
+            ["k1", "k2"],
+            lambda *k: ["x" for key in k],
+        )
+
+    def test_soft_invalidate_requires_expire_time_get_multi_second_call(self):
+        reg = self._region()
+        reg.get_or_create_multi(["k1", "k2"], lambda *k: ["x" for key in k])
+
+        reg.invalidate(hard=False)
+        assert_raises_message(
+            exception.DogpileCacheException,
+            "Non-None expiration time required for soft invalidation",
+            reg.get_or_create_multi,
+            ["k1", "k2"],
+            lambda *k: ["x" for key in k],
         )
 
     def test_should_cache_fn(self):
@@ -345,25 +455,26 @@ class RegionTest(TestCase):
 
         def creator():
             return values.pop(0)
-        should_cache_fn = lambda val: val in (1, 3)
+
+        should_cache_fn = lambda val: val in (1, 3)  # noqa
         ret = reg.get_or_create(
-            "some key", creator,
-            should_cache_fn=should_cache_fn)
+            "some key", creator, should_cache_fn=should_cache_fn
+        )
         eq_(ret, 1)
-        eq_(reg.backend._cache['some key'][0], 1)
-        time.sleep(.1)
+        eq_(reg.backend._cache["some key"][0], 1)
+        time.sleep(0.1)
         reg.invalidate()
         ret = reg.get_or_create(
-            "some key", creator,
-            should_cache_fn=should_cache_fn)
+            "some key", creator, should_cache_fn=should_cache_fn
+        )
         eq_(ret, 2)
-        eq_(reg.backend._cache['some key'][0], 1)
+        eq_(reg.backend._cache["some key"][0], 1)
         reg.invalidate()
         ret = reg.get_or_create(
-            "some key", creator,
-            should_cache_fn=should_cache_fn)
+            "some key", creator, should_cache_fn=should_cache_fn
+        )
         eq_(ret, 3)
-        eq_(reg.backend._cache['some key'][0], 3)
+        eq_(reg.backend._cache["some key"][0], 3)
 
     def test_should_cache_fn_multi(self):
         reg = self._region()
@@ -372,70 +483,67 @@ class RegionTest(TestCase):
         def creator(*keys):
             v = values.pop(0)
             return [v for k in keys]
-        should_cache_fn = lambda val: val in (1, 3)
+
+        should_cache_fn = lambda val: val in (1, 3)  # noqa
         ret = reg.get_or_create_multi(
-            [1, 2], creator,
-            should_cache_fn=should_cache_fn)
+            [1, 2], creator, should_cache_fn=should_cache_fn
+        )
         eq_(ret, [1, 1])
         eq_(reg.backend._cache[1][0], 1)
-        time.sleep(.1)
+        time.sleep(0.1)
         reg.invalidate()
         ret = reg.get_or_create_multi(
-            [1, 2], creator,
-            should_cache_fn=should_cache_fn)
+            [1, 2], creator, should_cache_fn=should_cache_fn
+        )
         eq_(ret, [2, 2])
         eq_(reg.backend._cache[1][0], 1)
-        time.sleep(.1)
+        time.sleep(0.1)
         reg.invalidate()
         ret = reg.get_or_create_multi(
-            [1, 2], creator,
-            should_cache_fn=should_cache_fn)
+            [1, 2], creator, should_cache_fn=should_cache_fn
+        )
         eq_(ret, [3, 3])
         eq_(reg.backend._cache[1][0], 3)
 
     def test_should_set_multiple_values(self):
         reg = self._region()
-        values = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'}
+        values = {"key1": "value1", "key2": "value2", "key3": "value3"}
         reg.set_multi(values)
-        eq_(values['key1'], reg.get('key1'))
-        eq_(values['key2'], reg.get('key2'))
-        eq_(values['key3'], reg.get('key3'))
+        eq_(values["key1"], reg.get("key1"))
+        eq_(values["key2"], reg.get("key2"))
+        eq_(values["key3"], reg.get("key3"))
 
     def test_should_get_multiple_values(self):
         reg = self._region()
-        values = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'}
+        values = {"key1": "value1", "key2": "value2", "key3": "value3"}
         reg.set_multi(values)
-        reg_values = reg.get_multi(['key1', 'key2', 'key3'])
-        eq_(
-            reg_values,
-            ["value1", "value2", "value3"]
-        )
+        reg_values = reg.get_multi(["key1", "key2", "key3"])
+        eq_(reg_values, ["value1", "value2", "value3"])
 
     def test_should_delete_multiple_values(self):
         reg = self._region()
-        values = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'}
+        values = {"key1": "value1", "key2": "value2", "key3": "value3"}
         reg.set_multi(values)
-        reg.delete_multi(['key2', 'key1000'])
-        eq_(values['key1'], reg.get('key1'))
-        eq_(NO_VALUE, reg.get('key2'))
-        eq_(values['key3'], reg.get('key3'))
+        reg.delete_multi(["key2", "key1000"])
+        eq_(values["key1"], reg.get("key1"))
+        eq_(NO_VALUE, reg.get("key2"))
+        eq_(values["key3"], reg.get("key3"))
 
 
 class ProxyRegionTest(RegionTest):
 
-    ''' This is exactly the same as the region test above, but it goes through
+    """This is exactly the same as the region test above, but it goes through
     a dummy proxy.  The purpose of this is to make sure the tests  still run
-    successfully even when there is a proxy '''
+    successfully even when there is a proxy"""
 
     class MockProxy(ProxyBackend):
-
         @property
         def _cache(self):
             return self.proxied._cache
 
     def _region(self, init_args={}, config_args={}, backend="mock"):
         reg = CacheRegion(**init_args)
-        config_args['wrap'] = [ProxyRegionTest.MockProxy]
+        config_args["wrap"] = [ProxyRegionTest.MockProxy]
         reg.configure(backend, **config_args)
         return reg
 
@@ -451,7 +559,6 @@ class CustomInvalidationStrategyTest(RegionTest):
     """
 
     class CustomInvalidationStrategy(RegionInvalidationStrategy):
-
         def __init__(self):
             self._soft_invalidated = None
             self._hard_invalidated = None
@@ -465,24 +572,27 @@ class CustomInvalidationStrategyTest(RegionTest):
                 self._hard_invalidated = None
 
         def is_invalidated(self, timestamp):
-            return ((self._soft_invalidated and
-                     timestamp < self._soft_invalidated) or
-                    (self._hard_invalidated and
-                     timestamp < self._hard_invalidated))
+            return (
+                self._soft_invalidated and timestamp < self._soft_invalidated
+            ) or (
+                self._hard_invalidated and timestamp < self._hard_invalidated
+            )
 
         def was_hard_invalidated(self):
             return bool(self._hard_invalidated)
 
         def is_hard_invalidated(self, timestamp):
-            return (self._hard_invalidated and
-                    timestamp < self._hard_invalidated)
+            return (
+                self._hard_invalidated and timestamp < self._hard_invalidated
+            )
 
         def was_soft_invalidated(self):
             return bool(self._soft_invalidated)
 
         def is_soft_invalidated(self, timestamp):
-            return (self._soft_invalidated and
-                    timestamp < self._soft_invalidated)
+            return (
+                self._soft_invalidated and timestamp < self._soft_invalidated
+            )
 
     def _region(self, init_args={}, config_args={}, backend="mock"):
         reg = CacheRegion(**init_args)
@@ -491,8 +601,141 @@ class CustomInvalidationStrategyTest(RegionTest):
         return reg
 
 
-class ProxyBackendTest(TestCase):
+class TestProxyValue(object):
+    def __init__(self, value):
+        self.value = value
 
+
+class MutexAPITest(TestCase):
+    def test_mutex_non_match(self):
+        assert not isinstance(5, CacheMutex)
+        assert not isinstance("some string", CacheMutex)
+
+        class Foo:
+            def release(self):
+                pass
+
+        assert not isinstance(Foo(), CacheMutex)
+
+        class Bar:
+            def acquire(self):
+                pass
+
+        assert not isinstance(Bar(), CacheMutex)
+
+    def test_mutex_match(self):
+        class Foo:
+            def acquire(self):
+                # note we are missing the "wait" parameter.  not sure
+                # how ABCs are supposed to efficiently test these
+                # things.
+                pass
+
+            def release(self):
+                pass
+
+        assert isinstance(Foo(), CacheMutex)
+
+
+class AsyncCreatorTest(TestCase):
+    def _fixture(self):
+        def async_creation_runner(cache, somekey, creator, mutex):
+            try:
+                value = creator()
+                cache.set(somekey, value)
+            finally:
+                mutex.release()
+
+        return mock.Mock(side_effect=async_creation_runner)
+
+    def test_get_or_create(self):
+        acr = self._fixture()
+        reg = CacheRegion(async_creation_runner=acr)
+        reg.configure("mock", expiration_time=0.2)
+
+        def some_value():
+            return "some value"
+
+        def some_new_value():
+            return "some new value"
+
+        eq_(reg.get_or_create("some key", some_value), "some value")
+        time.sleep(0.5)
+        eq_(reg.get_or_create("some key", some_new_value), "some value")
+        eq_(reg.get_or_create("some key", some_new_value), "some new value")
+        eq_(
+            acr.mock_calls,
+            [
+                mock.call(
+                    reg, "some key", some_new_value, reg._mutex("some key")
+                )
+            ],
+        )
+
+    def test_fn_decorator(self):
+        acr = self._fixture()
+        reg = CacheRegion(async_creation_runner=acr)
+        reg.configure("mock", expiration_time=5)
+
+        canary = mock.Mock()
+
+        @reg.cache_on_arguments()
+        def go(x, y):
+            canary(x, y)
+            return x + y
+
+        eq_(go(1, 2), 3)
+        eq_(go(1, 2), 3)
+
+        eq_(canary.mock_calls, [mock.call(1, 2)])
+
+        eq_(go(3, 4), 7)
+
+        eq_(canary.mock_calls, [mock.call(1, 2), mock.call(3, 4)])
+
+        reg.invalidate(hard=False)
+
+        eq_(go(1, 2), 3)
+
+        eq_(
+            canary.mock_calls,
+            [mock.call(1, 2), mock.call(3, 4), mock.call(1, 2)],
+        )
+
+        eq_(
+            acr.mock_calls,
+            [
+                mock.call(
+                    reg,
+                    "tests.cache.test_region:go|1 2",
+                    mock.ANY,
+                    reg._mutex("tests.cache.test_region:go|1 2"),
+                )
+            ],
+        )
+
+    def test_fn_decorator_with_kw(self):
+        acr = self._fixture()
+        reg = CacheRegion(async_creation_runner=acr)
+        reg.configure("mock", expiration_time=5)
+
+        @reg.cache_on_arguments()
+        def go(x, **kw):
+            return x
+
+        test_value = TestProxyValue("Decorator Test")
+        self.assertRaises(ValueError, go, x=1, foo=test_value)
+
+        @reg.cache_on_arguments()
+        def go2(x):
+            return x
+
+        # keywords that match positional names can be passed
+        result = go2(x=test_value)
+        self.assertTrue(isinstance(result, TestProxyValue))
+
+
+class ProxyBackendTest(TestCase):
     class GetCounterProxy(ProxyBackend):
         counter = 0
 
@@ -509,11 +752,12 @@ class ProxyBackendTest(TestCase):
 
     class UsedKeysProxy(ProxyBackend):
 
-        ''' Keep a counter of hose often we set a particular key'''
+        """ Keep a counter of hose often we set a particular key"""
 
         def __init__(self, *args, **kwargs):
             super(ProxyBackendTest.UsedKeysProxy, self).__init__(
-                *args, **kwargs)
+                *args, **kwargs
+            )
             self._key_count = defaultdict(lambda: 0)
 
         def setcount(self, key):
@@ -525,12 +769,13 @@ class ProxyBackendTest(TestCase):
 
     class NeverSetProxy(ProxyBackend):
 
-        ''' A totally contrived example of a Proxy that we pass arguments to.
-        Never set a key that matches never_set '''
+        """A totally contrived example of a Proxy that we pass arguments to.
+        Never set a key that matches never_set"""
 
         def __init__(self, never_set, *args, **kwargs):
             super(ProxyBackendTest.NeverSetProxy, self).__init__(
-                *args, **kwargs)
+                *args, **kwargs
+            )
             self.never_set = never_set
             self._key_count = defaultdict(lambda: 0)
 
@@ -554,8 +799,9 @@ class ProxyBackendTest(TestCase):
         return reg
 
     def test_cachedvalue_passed(self):
-        reg = self._region(config_args={"wrap": [
-            ProxyBackendTest.CanModifyCachedValueProxy]})
+        reg = self._region(
+            config_args={"wrap": [ProxyBackendTest.CanModifyCachedValueProxy]}
+        )
 
         reg.set("some key", "some value")
         eq_(reg.get("some key"), "some value")
@@ -565,9 +811,14 @@ class ProxyBackendTest(TestCase):
         # to the backend properly.  Test that methods not overridden
         # continue to work
 
-        reg = self._region(config_args={"wrap": [
-            ProxyBackendTest.GetCounterProxy,
-            ProxyBackendTest.SetCounterProxy]})
+        reg = self._region(
+            config_args={
+                "wrap": [
+                    ProxyBackendTest.GetCounterProxy,
+                    ProxyBackendTest.SetCounterProxy,
+                ]
+            }
+        )
         ProxyBackendTest.GetCounterProxy.counter = 0
         ProxyBackendTest.SetCounterProxy.counter = 0
 
@@ -601,21 +852,21 @@ class ProxyBackendTest(TestCase):
         reg_abc = self._region(config_args={"wrap": [proxy_abc]})
         for i in range(10):
             reg_num.set(i, True)
-            reg_abc.set(chr(ord('a') + i), True)
+            reg_abc.set(chr(ord("a") + i), True)
 
         for i in range(5):
             reg_num.set(i, True)
-            reg_abc.set(chr(ord('a') + i), True)
+            reg_abc.set(chr(ord("a") + i), True)
 
         # make sure proxy_num has the right counts per key
         eq_(proxy_num.setcount(1), 2)
         eq_(proxy_num.setcount(9), 1)
-        eq_(proxy_num.setcount('a'), 0)
+        eq_(proxy_num.setcount("a"), 0)
 
         # make sure proxy_abc has the right counts per key
-        eq_(proxy_abc.setcount('a'), 2)
-        eq_(proxy_abc.setcount('g'), 1)
-        eq_(proxy_abc.setcount('9'), 0)
+        eq_(proxy_abc.setcount("a"), 2)
+        eq_(proxy_abc.setcount("g"), 1)
+        eq_(proxy_abc.setcount("9"), 0)
 
     def test_argument_proxies(self):
         # Test that we can pass an argument to Proxy on creation
@@ -627,3 +878,130 @@ class ProxyBackendTest(TestCase):
         # make sure 1 was set, but 5 was not
         eq_(reg.get(5), NO_VALUE)
         eq_(reg.get(1), True)
+
+    def test_actual_backend_proxied(self):
+        # ensure that `reg.actual_backend` is the actual backend
+        # also ensure that `reg.backend` is a proxied backend
+        reg = self._region(
+            config_args={
+                "wrap": [
+                    ProxyBackendTest.GetCounterProxy,
+                    ProxyBackendTest.SetCounterProxy,
+                ]
+            }
+        )
+        assert isinstance(reg.backend, ProxyBackend)
+        assert isinstance(reg.actual_backend, CacheBackend)
+
+    def test_actual_backend_noproxy(self):
+        # ensure that `reg.actual_backend` is the actual backend
+        # also ensure that `reg.backend` is NOT a proxied backend
+        reg = self._region()
+        assert isinstance(reg.backend, CacheBackend)
+        assert isinstance(reg.actual_backend, CacheBackend)
+
+
+class LoggingTest(TestCase):
+    def _region(self, init_args={}, config_args={}, backend="mock"):
+        reg = CacheRegion(**init_args)
+        reg.configure(backend, **config_args)
+        return reg
+
+    def test_log_time(self):
+        reg = self._region()
+
+        times = [50, 55, 60]
+
+        def mock_time():
+            return times.pop(0)
+
+        with mock.patch("dogpile.cache.region.log") as mock_log, mock.patch(
+            "dogpile.cache.region.time", mock.Mock(time=mock_time)
+        ):
+            with reg._log_time(["foo", "bar", "bat"]):
+                pass
+
+        eq_(
+            mock_log.mock_calls,
+            [
+                mock.call.debug(
+                    "Cache value generated in %(seconds).3f "
+                    "seconds for key(s): %(keys)r",
+                    {
+                        "seconds": 5,
+                        "keys": util.repr_obj(["foo", "bar", "bat"]),
+                    },
+                )
+            ],
+        )
+
+    def test_repr_obj_truncated(self):
+
+        eq_(
+            repr(util.repr_obj(["some_big_long_name" for i in range(200)])),
+            "['some_big_long_name', 'some_big_long_name', "
+            "'some_big_long_name', 'some_big_long_name', 'some_big_long_name',"
+            " 'some_big_long_name', 'some_big_long_na ... "
+            "(4100 characters truncated) ... me_big_long_name', "
+            "'some_big_long_name', 'some_big_long_name', 'some_big_long_"
+            "name', 'some_big_long_name', 'some_big_long_name', "
+            "'some_big_long_name']",
+        )
+
+    def test_log_is_cache_miss(self):
+        reg = self._region()
+
+        with mock.patch("dogpile.cache.region.log") as mock_log:
+            is_(reg._is_cache_miss(NO_VALUE, "some key"), True)
+        eq_(
+            mock_log.mock_calls,
+            [mock.call.debug("No value present for key: %r", "some key")],
+        )
+
+    def test_log_is_value_version_miss(self):
+
+        reg = self._region()
+        inv = mock.Mock(is_hard_invalidated=lambda val: True)
+        with mock.patch(
+            "dogpile.cache.region.log"
+        ) as mock_log, mock.patch.object(reg, "region_invalidator", inv):
+            is_(
+                reg._is_cache_miss(
+                    CachedValue(
+                        "some value", {"v": value_version - 5, "ct": 500}
+                    ),
+                    "some key",
+                ),
+                True,
+            )
+        eq_(
+            mock_log.mock_calls,
+            [
+                mock.call.debug(
+                    "Dogpile version update for key: %r", "some key"
+                )
+            ],
+        )
+
+    def test_log_is_hard_invalidated(self):
+
+        reg = self._region()
+        inv = mock.Mock(is_hard_invalidated=lambda val: True)
+        with mock.patch(
+            "dogpile.cache.region.log"
+        ) as mock_log, mock.patch.object(reg, "region_invalidator", inv):
+            is_(
+                reg._is_cache_miss(
+                    CachedValue("some value", {"v": value_version, "ct": 500}),
+                    "some key",
+                ),
+                True,
+            )
+        eq_(
+            mock_log.mock_calls,
+            [
+                mock.call.debug(
+                    "Hard invalidation detected for key: %r", "some key"
+                )
+            ],
+        )
